import os
from pathlib import Path

from django.conf import settings
from django.core.exceptions import ValidationError
from django.test import SimpleTestCase
from django.utils import six
from django.test.utils import override_settings

import django_tools
from django_tools.validators import URLValidator2, ExistingDirValidator


@override_settings(DEBUG=False)
class TestExistingDirValidator(SimpleTestCase):
    @classmethod
    def setUpClass(cls):
        super(TestExistingDirValidator, cls).setUpClass()
        cls.media_root_validator = ExistingDirValidator()

    def test_default_media_root(self):
        self.media_root_validator(settings.MEDIA_ROOT)
        try:
            self.media_root_validator("does/not/exist")
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "Directory doesn't exist!")

    @override_settings(DEBUG=True)
    def test_debug_message(self):
        path = str(Path(settings.MEDIA_ROOT, "does/not/exist"))
        print(path)
        try:
            self.media_root_validator(path)
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "Directory '%s' doesn't exist!" % path)

    def test_existing_dirs(self):
        BASE_PATH=os.path.abspath(os.path.dirname(django_tools.__file__))
        validator = ExistingDirValidator(BASE_PATH)
        for root, dirs, files in os.walk(BASE_PATH):
            for dir in dirs:
                path = os.path.join(root, dir)
                validator(path)

    def test_not_in_media_root1(self):
        try:
            self.media_root_validator("../")
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "Directory is not in base path!")

    def test_not_in_media_root2(self):
        try:
            self.media_root_validator("//")
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "Directory is not in base path!")

    def test_directory_traversal_attack_encodings(self):
        parts = (
            # URI encoded directory traversal:
            "%2e%2e%2f", # ../
            "%2e%2e/", # ../
            "..%2f", # ../
            "%2e%2e%5c", # ..\

            # Unicode / UTF-8 encoded directory traversal:
            "..%c1%1c", # ../
            "..%c0%af", # ..\
            "%c0%ae", # .
        )
        for part in parts:
            try:
                self.media_root_validator(part)
            except ValidationError as err:
                self.assertEqual(six.text_type(err.message), "Directory doesn't exist!")



class TestUrlValidator(SimpleTestCase):
    def test_scheme_without_netloc(self):
        try:
            URLValidator2(allow_all_schemes=True, allow_netloc=False)
        except AssertionError as err:
            self.assertEqual(str(err), "Can't allow schemes without netloc!")

    def test_no_allow_all_schemes(self):
        try:
            URLValidator2(allow_schemes=("http","ftp"), allow_all_schemes=True)
        except Warning as err:
            self.assertEqual(str(err), "allow_schemes would be ignored, while allow_all_schemes==True!")

    def test_not_start_with_allow_all_schemes(self):
        URLValidator2(allow_schemes=("svn",))("svn://domain.test")
        try:
            URLValidator2(allow_schemes=("http","ftp"))("svn://domain.test")
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "The URL doesn't start with a allowed scheme.")

    def test_allow_query(self):
        validator = URLValidator2(allow_query=False)
        validator("http://www.domain.test/without/query/")
        try:
            validator("http://www.domain.test/with/?query")
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "Enter a valid URL without a query.")

    def test_allow_fragment(self):
        validator = URLValidator2(allow_fragment=False)
        validator("http://www.domain.test/without/fragment/")
        try:
            validator("http://www.domain.test/with/a/#fragment")
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "Enter a valid URL without a fragment.")

    def test_only_local_path1(self):
        validator = URLValidator2(allow_schemes=None, allow_netloc=False)
        validator("/path/?query#fragment")
        try:
            validator("http://domain.test/path/?query#fragment")
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "Please enter a local URL (without protocol/domain).")

    def test_only_local_path2(self):
        """
        **Note:** Validating the network location (netloc):
        Following the syntax specifications in RFC 1808, urlparse recognizes a
        netloc only if it is properly introduced by '//'. Otherwise the input is
        presumed to be a relative URL and thus to start with a path component.
        See: http://docs.python.org/library/urlparse.html#urlparse.urlparse
        """
        validator = URLValidator2(allow_schemes=None, allow_netloc=False)
        validator("www.pylucid.org/path?query#fragment")
        try:
            validator("//www.pylucid.org/path?query#fragment")
        except ValidationError as err:
            self.assertEqual(six.text_type(err.message), "Please enter a local URL (without protocol/domain).")



# coding: utf-8

"""
    permission_info manage command


    setup:

        INSTALLED_APPS = [
            ...
            "django_tools",
            ...
        ]


    usage:

        $ ./manage.py permission_info
        $ ./manage.py permission_info <username>


    :copyleft: 2017 by the django-tools team, see AUTHORS for more details.
    :created: 2017 by Jens Diemer
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.core.management.base import BaseCommand

# https://github.com/jedie/django-tools
from django_tools.permissions import pformat_permission


class Command(BaseCommand):
    help = (
        "Display effective user permissions"
        " in the same format as user.has_perm() argument:"
        " <appname>.<codename>"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "user", nargs="?", default=None,
            help="Username"
        )

    def list_usernames(self):
        self.stdout.write("\nAll existing users are:\n")
        UserModel = get_user_model()
        qs = UserModel.objects.all()
        user_count = qs.count()
        usernames = qs.values_list("username", flat=True).order_by("username")
        self.stdout.write(", ".join(usernames))
        self.stdout.write("(%i users)" % user_count)

    def username_error(self):
        self.stderr.write("No username given!")
        self.list_usernames()
        self.stdout.write("")

    def handle(self, *args, **options):
        self.stdout.write("")
        self.stdout.write("_"*79)
        self.stdout.write(self.help)
        self.stdout.write("")

        username = options.get("user")

        if not username:
            self.username_error()
            return

        self.stdout.write("All permissions for user '%s':" % username)

        UserModel = get_user_model()
        try:
            user = UserModel.objects.get(username=username)
        except UserModel.DoesNotExist as err:
            self.stderr.write("Username %r doesn't exists: %s" % (username, err))
            self.list_usernames()
            return

        for attr_name in ("is_active", "is_staff", "is_superuser"):
            info = "yes" if getattr(user, attr_name) else "no"
            self.stdout.write("\t%-13s: %s" % (attr_name, info))

        if user.is_superuser:
            self.stdout.write("Don't list permissions: Superusers has all permissions ;)")
            return

        from django.contrib import auth
        all_permissions = set()

        for backend in auth.get_backends():
            if hasattr(backend, "get_all_permissions"):
                all_permissions.update(backend.get_all_permissions(user))

        seen_permissions = set()

        qs = Permission.objects.all().order_by("content_type__app_label", "content_type__model", "codename")
        for permission in qs.iterator():
            perm_name = pformat_permission(permission)
            seen_permissions.add(perm_name)

            contains = "[*]" if user.has_perm(perm_name) else "[ ]"
            self.stdout.write("%s %s" % (contains, perm_name))

        # We should never miss a permission, but we can still try it out:
        not_seen = all_permissions - seen_permissions
        for perm_name in sorted(not_seen):
            self.stderr.write("missing: %s" % perm_name)



# coding: utf-8

"""
    Put this into your settings:
    --------------------------------------------------------------------------
        MIDDLEWARE_CLASSES = (
            ...
            'django_tools.middlewares.TracebackLogMiddleware.TracebackLogMiddleware',
            ...
        )
    --------------------------------------------------------------------------

    :copyleft: 2016 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import unicode_literals

import logging

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object  # fallback for Django < 1.10


class TracebackLogMiddleware(MiddlewareMixin):

    def process_exception(self, request, exception):
        logging.exception('Exception on url: %s' % request.path)




# coding: utf-8

"""
    Mockups
    ~~~~~~~
"""


import tempfile

from django.core.files import File as DjangoFile

from PIL import Image, ImageDraw, ImageFont

try:
    from filer.models import Image as FilerImage
except ImportError as err:
    FilerImage = None
    FILER_IMPORT_ERROR = err


DUMMY_TEXT = """Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean commodo ligula
eget dolor. Aenean massa. Cum sociis natoque penatibus et magnis dis parturient
montes, nascetur ridiculus mus. Donec quam felis, ultricies nec, pellentesque
eu, pretium quis, sem. Nulla consequat massa quis enim. Donec pede justo,
fringilla vel, aliquet nec, vulputate eget, arcu. In enim justo, rhoncus ut,
imperdiet a, venenatis vitae, justo. Nullam dictum felis eu pede mollis pretium.

Integer tincidunt. Cras dapibus. Vivamus elementum semper nisi. Aenean vulputate
eleifend tellus. Aenean leo ligula, porttitor eu, consequat vitae, eleifend ac,
enim. Aliquam lorem ante, dapibus in, viverra quis, feugiat a, tellus. Phasellus
viverra nulla ut metus varius laoreet. Quisque rutrum. Aenean imperdiet. Etiam
ultricies nisi vel augue. Curabitur ullamcorper ultricies nisi. Nam eget dui.

Etiam rhoncus. Maecenas tempus, tellus eget condimentum rhoncus, sem quam semper
libero, sit amet adipiscing sem neque sed ipsum. Nam quam nunc, blandit vel,
luctus pulvinar, hendrerit id, lorem. Maecenas nec odio et ante tincidunt
tempus. Donec vitae sapien ut libero venenatis faucibus. Nullam quis ante. Etiam
sit amet orci eget eros faucibus tincidunt. Duis leo. Sed fringilla mauris sit
amet nibh. Donec sodales sagittis magna. Sed consequat, leo eget bibendum
sodales, augue velit cursus nunc"""


def create_filer_image(pil_image, user):
    """
    Create from a PIL image a filer.models.Image() instance
    """
    if FilerImage is None:
        # Django-Filer is not available: raise the origin error
        raise ImportError(FILER_IMPORT_ERROR)

    file_obj = DjangoFile(pil_image, name=pil_image.name)
    image = FilerImage.objects.create(owner=user, original_filename=pil_image.name, file=file_obj, folder=None)
    return image


class ImageDummy:
    new_image_color = "black"
    text_color = "#ffffff"
    text_align = "center"
    temp_prefix = "dummy_"
    format = "png"

    def __init__(self, width, height):
        self.width = width
        self.height = height

    def fill_image(self, image):
        """
        Fill a PIL image with a colorful gradient.

        Overwrite with e.g.:
            self.draw_centered_text(
                image,
                text="(dummy picture)",
                font_size_factor=16,
                truetype_font="DejaVuSansMono.ttf"
            )
        """
        pixel_map = image.load()
        for i in range(self.width):
            for j in range(self.height):
                pixel_map[i, j] = (i, j, 1)

    def draw_centered_text(self, image, text, color="#000000", size_factor=16, truetype=None):
        """
        Draw the given >text< centered on the given >image<
        Maybe useful for self.fill_image()

        FIXME: Why in hell is it so complicated to draw a centered text with pillow?!?

        :param image: PIL instance, e.g.: Image.new()
        :param text: The text to draw
        :param color: Text color (font fill color)
        :param size_factor: used to calc the font size by image size
        :param truetype: for ImageFont.truetype.font, e.g.: "DejaVuSansMono.ttf"
        :return: None
        """
        draw = ImageDraw.Draw(image)

        font_size = min([self.width, self.height])
        font_size = int(font_size / size_factor)

        if truetype is not None:
            font = ImageFont.truetype(font=truetype, size=font_size)
            split_character = "\n" if isinstance(text, str) else b"\n"
            lines = text.split(split_character)
            max_width = 0
            widths = []
            for line in lines:
                line_width, line_height = font.getsize(line)
                widths.append(line_width)
                max_width = max(max_width, line_width)

            line_width, line_height = font.getsize(text)
            left = int((self.width - max_width) / 2)
            top = int((self.height - line_height) / 2)
        else:
            font = None
            left = int((self.width) / 2)
            top = int((self.height) / 2)

        draw.multiline_text(xy=(left, top), text=text, fill=color, font=font, align="center")

    def create_pil_image(self):
        """
        return a 'filled' PIL image
        """
        image = Image.new("RGB", (self.width, self.height), self.new_image_color)
        self.fill_image(image)
        return image

    def create_info_image(self, text):
        """
        return a 'filled' PIL image with >text<
        """
        image = self.create_pil_image()
        draw = ImageDraw.Draw(image)
        draw.multiline_text((10, 10), text, fill=self.text_color, align=self.text_align)
        return image

    def create_temp_filer_info_image(self, text, user):
        """
        Fill a PIL image with a colorful gradient,
        draw the given >text< on it
        and return a filer.models.Image() instance.
        """
        if FilerImage is None:
            # Django-Filer is not available: raise the origin error
            raise ImportError(FILER_IMPORT_ERROR)

        f = tempfile.NamedTemporaryFile(prefix=self.temp_prefix, suffix=".%s" % self.format)
        image = self.create_info_image(text)
        image.save(f, format=self.format)
        filer_image = create_filer_image(f, user)
        return filer_image
